use std::rc::{Rc, Weak};
use std::cell::RefCell;
use r9_macro::SharedFromSelf;
use r9_macro_derive::SharedFromSelf;
use r9_macro_derive::Singleton;
use r9::Queue;
use f9::app::App;
use f9::app::HttpContext;
use crate::listener::WsListener;
use crate::upstream::UpStreamMgr;
use bytes::BytesMut;
//use crate::ss;

pub struct DownStreamPack {
    data: bytes::BytesMut,
    entry: Rc::<RefCell::<r9::ListHead<DownStreamPack>>>,
}

#[derive(SharedFromSelf)]
#[derive(Singleton)]
pub struct UserApp {
    _self_wp: Weak::<RefCell::<Self>>,
    net_msg_queue: Rc::<RefCell::<r9::Queue<DownStreamPack>>>,
}

impl f9::app::UserApp for UserApp {

    fn get_pkg_name(&self) -> String {
        return "statserver".to_string();
    }

    fn init(&mut self) {
        App::instance().borrow_mut().add_http_handler(
            "Ops$selfChecking".to_string(),
            Rc::new(RefCell::new(
                    move |ctx: &mut HttpContext| {
                        ctx.rsp = "Ops$selfChecking".to_string();
                        App::instance().borrow_mut().add_http_handler(
                            "Ops$selfChecking3".to_string(),
                            Rc::new(RefCell::new(
                                move |ctx: &mut HttpContext| {
                                    ctx.rsp = "Ops$selfChecking3".to_string();
                                }
                            )),
                        );
                    }
                )
            ),
        );
        WsListener::instance().borrow_mut().init();
        UpStreamMgr::instance().borrow_mut().init();
    }

    fn update(&mut self) {
        self.net_msg_queue.borrow_mut().fetch();
        if self.net_msg_queue.borrow_mut().empty() {

        }
    }

    fn uninit(&mut self) {
        WsListener::instance().borrow_mut().uninit();
    }

    fn get_http_listen_port(&self) -> i32 {
        return 8080;
    }

}

impl UserApp {

    pub fn add_down_stream_pack(&mut self, data: bytes::BytesMut) {
        let node = Rc::new(RefCell::new(DownStreamPack{
            data: data,
            entry: r9::ListHead::< crate::app::user_app::DownStreamPack>::new_node(Default::default()),
        }));
        node.borrow_mut().entry = r9::ListHead::< crate::app::user_app::DownStreamPack>::new_node(Rc::downgrade(&node));
        self.net_msg_queue.borrow_mut().push(&node.borrow_mut().entry);
    }

    pub fn new() -> Self {
        let p = UserApp{
            _self_wp: Default::default(),
            net_msg_queue: r9::Queue::<DownStreamPack>::new(),
        };
        return p;
    }

}
